<?php

namespace Slovocast;

use Slim\App;
use Slim\Factory\AppFactory;

use League\Config\Configuration;

use DI\Container;
use DI\ContainerBuilder;

use Slim\Views\Twig;
use Slim\Views\TwigMiddleware;

use Slovocast\Routes;
use Slovocast\Configuration\SiteInformationSchema;
use Slovocast\Configuration\DatabaseConnectionSchema;

use Psr\Log\LoggerInterface;
use Monolog\Logger;
use Monolog\Handler\StreamHandler;
use Monolog\Level;

class Bootstrap
{
    /**
     * Pulls out all the configuration schemas and configuration values. 
     *
     * @return Configuration
     */
    protected static function initConfig(): Configuration
    {
        $config = new Configuration();

        // set all configuration details
        $config->addSchema('site', SiteInformationSchema::getSchema());
        $config->addSchema('database', DatabaseConnectionSchema::getSchema());

        $config->merge([
            'site' => [
                'name' => "Slovocast",
                'description' => "A no-bullshit podcast hosting platform.",
                'tags' => [
                    'podcast', 'hosting', 'indie', 'independent', 'easy'
                ]
            ],
        ]);

        return $config;
    }

    /**
     * Set up the Container with all of its definitions, as well as
     * initialization of configuration.
     *
     * @return Container
     */
    protected static function initContainer(): Container 
    {
        $containerBuilder = new ContainerBuilder();

        $config = self::initConfig();
        $containerBuilder->addDefinitions([
            'config' => $config,
            // more DI stuff
            'flash' => function () {
                $messages = [];
                return new \Slim\Flash\Messages($messages);
            },
            LoggerInterface::class => function() {
                $logger = new Logger();
                $logger->pushHandler(
                    new StreamHandler(__DIR__ . '/var/logs', Level::Warning)
                );
                return $logger;
            }
        ]);

        return $containerBuilder->build();
    }

    /**
     * Tasking the instaniated Application, sets up all the routes for the 
     * application.
     *
     * @param App $app
     * @return void
     */
    protected static function establishRoutes(App $app): void
    {
        Routes::init($app);
    }

    /**
     * Taking an instantiated application, sets up all the middleware required
     * for the application to run.
     *
     * @param App $app
     */
    protected static function establishMiddleware(App $app): void
    {
        /**
         * @var Configuration
         */
        $config = $app->getContainer()->get('config');
        $app->addErrorMiddleware(true, true, true);

        // Twig
        $templatePath = __DIR__ . "/../templates";
        $templateCache = false;
        $twig = Twig::create($templatePath, [ 'cache' => $templateCache ]);

        // Add the global variables
        $twig->getEnvironment()
             ->addGlobal('site_name', $config->get('site.name'));
        $twig->getEnvironment()
             ->addGlobal('site_description', $config->get('site.description'));
        $app->add(TwigMiddleware::create($app, $twig));

        // Flash Messages
        $app->add(function ($req, $next) {
            if (session_status() !== PHP_SESSION_ACTIVE) {
                session_start();
            }

            /** @var $this ContainerInterface */
            $this->get('flash')->__construct($_SESSION);

            return $next->handle($req);
        });
    }

    /**
     * Instantiates the application.
     *
     * @return App
     */
    public static function init(): App
    {
        $container = self::initContainer();
        $app = AppFactory::createFromContainer($container);
        self::establishMiddleware($app);
        self::establishRoutes($app);
        return $app;
    }
}
